#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the THUCNews dataset."""

from .loader import THUCNews

__all__ = ["THUCNews"]
