#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Coin Image dataset."""

from .loader import CoinImage

__all__ = ["CoinImage"]
